"""Top level of GFF parsing providing shortcuts for useful classes.
"""
from BCBio.GFF.GFFParser import GFFParser, DiscoGFFParser, GFFExaminer, parse, parse_simple
from BCBio.GFF.GFFOutput import GFF3Writer, write

__version__ = "0.6.9"
